/*
Cory McKay

ComponentLabeler Class

- Objects of this class are created to estimate the number of times a binary image should be shrunk
  and expanded to remove noise or to generate labeled maps of objects.
- A new ComponentLabeler object should be created for each new image to be checked.
- The getEstimatedWhiteGrowths and the getEstimatedBlackGrowths methods can be called to get
  the estimated number of expansions/shrinkings that will be needed to remove black and white noise
  respectively.
- The getRegionMap method returns a MapCountObject containing a labelled map of the objects
  in the image and a count of the pixels with each label.
- This is done by finding the number of pixels in the largest 8-connected object of the appropriate
  colour in the image that is made up of 25 or fewer pixels. This limit is chosen because objects
  bigger than this might be more than just noise and because they are large enough that the effect of
  eliminating them through expanding and shrinking might distort the image too much. The
  square root - 1 of the size of the result is then found, since this is an approximate measure of
  how many expansions will be needed to eliminate this object. This value is returned as the number
  of necessary expansions.
*/

class  ComponentLabeler
{
	// Fields
	private char[][] pixels; // represents a binary image. 'b' represents a black pixel, 'w' represents
	                         //    a white pixel
	private int[][] map; // a map of objects that have been labelled. The indices correspond to those
	                     //    of the pixels array. The pixels belonging to a connected object (using
				         //    8-neighbours) are given the same integer value. The first object is given
				         //    the number 1. Pixels belonging to the background are given the number 0.
	private int current_region; // the number (used to mark map) of the object that is currently being
	                            //    labelled
	private int width; // the width of the image
	private int height; // the height of the image
	private boolean change_made; // whether or not a change has been made while updating the map of objects


	/***************************************************************************/


	public ComponentLabeler(char[][] pix, int w, int h)
	// Constructor.
	// Takes in the pixels array and its width and height.
	// Sets the current_region to 1.
	{
		pixels = pix;
		width = w;
		height = h;
		current_region = 1;
		map = new int[width][height];
		change_made = true;
	}


	/***************************************************************************/


	public int getEstimatedWhiteGrowths()
	// Return the estimated number of times that white objects should be expanded and
	// then shrunk to eliminate noise made up of black pixels. This is done by finding
	// the number of pixels in the largest 8-connected black object in the image that is
	// made up of 25 or fewer pixels. This limit is chosen because objects bigger than this
	// might be more than just noise and because they are large enough that the effect of
	// eliminating them through expanding and shrinking might distort the image too much.
	// The square root - 1 of the size of the result is then found, since this is an
	// approximate measure of how many expansions will be needed to eliminate this object.
	// This value is returned as the number of necessary expansions.
	{
		int number_expansions = 0; // the number of expansions needed to eliminate noise
		int rough_dimension = 0; // the rough dimension of an object
		int[] pixel_counts = getPixelCounts('b'); // the number of pixels in each object
		                                          // note: pixel_counts[0] holds the number
		                                          // of pixels in the background


		for (int i = 1; i < pixel_counts.length; i++) // i starts at 1 becasue must ignore background
		{
			if (pixel_counts[i] <= 25 && pixel_counts[i] != 0)
			{
				rough_dimension = (int) Math.sqrt( (double) pixel_counts[i] ) - 1;
				if (rough_dimension > number_expansions)
					number_expansions = rough_dimension;
			}
		}


		return number_expansions;
	}


	/***************************************************************************/


	public int getEstimatedBlackGrowths()
	// Return the estimated number of times that black objects should be expanded and
	// then shrunk to eliminate noise made up of white pixels. This is done by finding
	// the number of pixels in the largest 8-connected white object in the image that is
	// made up of 25 or fewer pixels. This limit is chosen because objects bigger than this
	// might be more than just noise and because they are large enough that the effect of
	// eliminating them through expanding and shrinking might distort the image too much.
	// The square root - 1 of the size of the result is then found, since this is an
	// approximate measure of how many expansions will be needed to eliminate this object.
	// This value is returned as the number of necessary expansions.
	{
		int number_expansions = 0; // the number of expansions needed to eliminate noise
		int rough_dimension = 0; // the rough dimension of an object
		int[] pixel_counts = getPixelCounts('w'); // the number of pixels in each object
		                                          // note: pixel_counts[0] holds the number
		                                          // of pixels in the background


		for (int i = 1; i < pixel_counts.length; i++) // i starts at 1 becasue must ignore background
		{
			if (pixel_counts[i] <= 25 && pixel_counts[i] != 0)
			{
				rough_dimension = (int) Math.sqrt( (double) pixel_counts[i] ) - 1;
				if (rough_dimension > number_expansions)
					number_expansions = rough_dimension;
			}
		}


		return number_expansions;
	}


	/***************************************************************************/


	public MapCountObject getRegionMap(char object_colour)
	// Returns a MapCountObject containing a labelled map of the objects
	// in the image and a count of the pixels with each label
	{
		int[] pixel_counts = getPixelCounts(object_colour); // the number of pixels in each object
						                                    // note: pixel_counts[0] holds the number
		                                                    // of pixels in the background
		return (new MapCountObject(map, pixel_counts));
	}


	/***************************************************************************/


	private int[] getPixelCounts(char object_colour)
	// Return an array with the number of pixels in each 8-connected object of the
	// colour object_colour (where 'w' is white and 'b' is black). The number at the
	// indice 0 is the number of pixels in the background (i.e. the number of pixels
	// of the colour opposite to object_colour). The numbers at the other indices
	// are the number of pixels in each object.
	// NOTE: due to the way objects are detected, some elements of the array may have
	// values of zero. This is not an error, it is due to the creation of false dummy
	// objects while linking pixels belonging to the same object together.
	{
		// Set up map to contain the information about each object
		labelMap(object_colour);


		// Find the number of objects found
		int max_index = 0;
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (map[x][y] > max_index)
					max_index = map[x][y];
			}
		}


		// Find and return the count of pixels in each object
		int[] number_pixels_per_object = new int[ max_index + 1 ];
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				number_pixels_per_object[ map[x][y] ]++;
			}
		}
		return number_pixels_per_object;
	}


	/***************************************************************************/


	private void labelMap(char object_colour)
	// Assigns the pixels of each of the 8-connected objects in the image an integer number
	// that corresponds to the number of the object. Objects are considered to be black and
	// the background is white if object_colour is 'b', and objects are considered to be white
	// and the background is black if object_colour is 'w'. The labelling is stored on the
	// map array.
	{
		// Reset the current_region and initialize the map array
		current_region = 1;
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				map[x][y] = 0;
			}
		}


		// Go through the image and assign integer values to each object
		change_made = true;
		while (change_made == true)
		{
			change_made = false;
			for (int y = 0; y < height; y++)
			{
				for (int x = 0; x < width; x++)
				{
					if ( pixels[x][y] == object_colour)
						markComponent(x, y);
				}
			}
		}
	}


	/***************************************************************************/


	private void markComponent(int x, int y)
	// Looks at the 8-neighbours of the point (x,y). If it has an object as a neighbour,
	// then assigns its object number to map[x][y]. If there is no neighbouring object,
	// then increments current_region and assigns the new object number to map[x][y].
	{
		int adjacent_region = 999999; // the object number assigned to a neighbouring pixels
		                              // 999999 is just a dumby number
		int original_value = map[x][y]; // the value of the pixel before it is altered

		// Examine 8-neighbours, being careful not to exceed the limits of the array,
		// to see if any of them have been labelled as an object
		if ( x != 0 && y != 0 )
		{
			if (map[x-1][y-1] != 0 && map[x-1][y-1] < adjacent_region)
				adjacent_region = map[x-1][y-1];
		}
		if ( x != 0 )
		{
			if (map[x-1][y] != 0 && map[x-1][y] < adjacent_region)
				adjacent_region = map[x-1][y];
		}
		if ( x != 0 && y != height - 1 )
		{
			if (map[x-1][y+1] != 0 && map[x-1][y+1] < adjacent_region)
				adjacent_region = map[x-1][y+1];
		}
		if ( y != 0 )
		{
			if (map[x][y-1] != 0 && map[x][y-1] < adjacent_region)
				adjacent_region = map[x][y-1];
		}
		if ( y != height - 1 )
		{
			if (map[x][y+1] != 0 && map[x][y+1] < adjacent_region)
				adjacent_region = map[x][y+1];
		}
		if ( x != width - 1 && y != 0 )
		{
			if (map[x+1][y-1] != 0 && map[x+1][y-1] < adjacent_region)
				adjacent_region = map[x+1][y-1];
		}
		if ( x != width - 1 )
		{
			if (map[x+1][y] != 0 && map[x+1][y] < adjacent_region)
				adjacent_region = map[x+1][y];
		}
		if ( x != width - 1 && y != height - 1 )
		{
			if (map[x+1][y+1] != 0 && map[x+1][y+1] < adjacent_region)
				adjacent_region = map[x+1][y+1];
		}


		// If the given pixel is adjacent to another object, assign it the number of that
		// object in the map array. If it is not adjacent to another object, assign it a new
		// object number and label it with this number on map.
		if (adjacent_region < 999999)
			map[x][y] = adjacent_region;
		if (map[x][y] == 0)
		{
			map[x][y] = current_region;
			current_region++;
		}


		// Check to see if any changes have been made
		if (map[x][y] != original_value)
			change_made = true;
	}
}


	/***************************************************************************/


/*  THE FOLLOWING METHOD WAS USED FOR TESTING PURPOSES

	public int[][] getMap(char object_colour)
	{
		labelMap(object_colour);
		return map;
	}
*/


/*  THE FOLLOWING TWO METHODS WERE ORIGINALLY USED TO LABEL OBJECTS RECURSIVELY. THIS APPROACH
    WAS DISCONTINUED, HOWEVER, BECAUSE THE RECURSION RESULTED IN A STACK OVERFLOW FOR LARGE
    OBJECTS

	private void labelMap(char object_colour)
	// Assigns the pixels of each of the 8-connected objects in the image an integer number
	// that corresponds to the number of the object. Objects are considered to be black and
	// the background is white if object_colour is 'b', and objects are considered to be white
	// and the background is black if object_colour is 'w'. The labelling is stored on the
	// map array.
	{
		// Reset the current_region and initialize the map array
		current_region = 1;
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				map[x][y] = 0;
			}
		}


		// Go through the image and assign integer values to each object
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if ( pixels[x][y] == object_colour &&  map[x][y] == 0 )
				{
					markComponent(x, y, object_colour);
					current_region++;
				}
			}
		}
	}

	private void markComponent(int x, int y, char object_colour)
	// Assign an integer to all pixels of the colour object_colour ('b' for black
	// and 'w' for white) that are 8-connected to the pixel at coordinates (x, y).
	// This is done by recursively examining the neighbours of x and y. These integers
	// are stored in the map array.
	{
		// Mark the current pixel
		map[x][y] = current_region;


		// Examine 8-neighbours, being careful not to exceed the limits of the array
		if ( x != 0 && y != 0 )
		{
			if (pixels[x-1][y-1] == object_colour && map[x-1][y-1] == 0)
				markComponent(x-1, y-1, object_colour);
		}
		if ( x != 0 )
		{
			if (pixels[x-1][y] == object_colour && map[x-1][y] == 0)
				markComponent(x-1, y, object_colour);
		}
		if ( x != 0 && y != height - 1 )
		{
			if (pixels[x-1][y+1] == object_colour && map[x-1][y+1] == 0)
				markComponent(x-1, y+1, object_colour);
		}
		if ( y != 0 )
		{
			if (pixels[x][y-1] == object_colour && map[x][y-1] == 0)
				markComponent(x, y-1, object_colour);
		}
		if ( y != height - 1 )
		{
			if (pixels[x][y+1] == object_colour && map[x][y+1] == 0)
				markComponent(x, y+1, object_colour);
		}
		if ( x != width - 1 && y != 0 )
		{
			if (pixels[x+1][y-1] == object_colour && map[x+1][y-1] == 0)
				markComponent(x+1, y-1, object_colour);
		}
		if ( x != width - 1 )
		{
			if (pixels[x+1][y] == object_colour && map[x+1][y] == 0)
				markComponent(x+1, y, object_colour);
		}
		if ( x != width - 1 && y != height - 1 )
		{
			if (pixels[x+1][y+1] == object_colour && map[x+1][y+1] == 0)
				markComponent(x+1, y+1, object_colour);
		}
	}
*/