/*
Cory McKay

EdgeDetector Class

- This class is used to detect edges in images, using the Sobel operator.
- The getFragments method can be called to uses the Sobel Operator to detect edges in the image with the
  brightness levels corresponding to the brightness_map 2-D array. It populates the fragment_magnitude
  and fragment_orientation 2-D arrays with the magnitude and orientation of detected edges. This method
  also offers the option of applying a mean filter before applying the operator. This method returns
  a boolean map of the image that contains true values at the coordinates where a significant
  edge fragment was detected.
*/

class  EdgeDetector
{
	// Fields
	private int width;           // the width of image
	private int height;          // the height of image
	private int[][] brightness_map; // a map of brightness levels in the histogram, from 0 to 255
	private double[][] fragment_magnitude; // a map of the magnitude of edge fragments. Is -1 when empty.
	private double[][] fragment_orientation; // a map of the orientation of edge fragments. Is -1 when empty.


	/***************************************************************************/


	public EdgeDetector (int w, int h, int[][] map)
	// Constructor
	// Map is a 2-D array of brightness levels in an image (from 0 to 255) and w an h are the image's
	// width and height
	// Initializes fragment_magnitude and fragment_orientation to empty
	{
		width = w;
		height = h;
		brightness_map = map;

		fragment_magnitude = new double[width][height];
		fragment_orientation = new double[width][height];
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				fragment_magnitude[x][y] = -1;
				fragment_orientation[x][y] = -1;
			}
		}
	}


	/***************************************************************************/


	public boolean[][] getFragments(boolean want_apply_filter)
	// Applies a mean filter to the brightness data in brightness map and then uses a Sobel Operator to detect edges.
	// Stores the edge strengths in fragment_magnitude and the edge orientations in fragment_orientation.
	// Returns a map that contains true in cells where the edge strength is greater than 80. This is considered a strong
	// enough edge fragment to be considered acceptable, and this threshold was arrived at experimentally.
	// want_apply_filter should be false if it is not desired that the mean filter be applied. It should be
	// false otherwise.
	{
		boolean edge_fragment_present[][] = new boolean[width][height]; // a map of detected edges
		int x_sobel; // the results of the x-Sobel on a given pixel
		int y_sobel; // the results of the y-Sobel on a given pixel


		// Apply a mean filter if requested
		if (want_apply_filter == true)
			apply_mean_filter();


		// Calculate the edge strength and orientation for each pixel
		for (int y = 1; y < height - 1; y++)
		{
			for (int x = 1; x < width - 1; x++)
			{
				x_sobel = (-1 * brightness_map[x-1][y-1]) + (1 * brightness_map[x+1][y-1]) +
					      (-2 * brightness_map[x-1][y]) + (2 * brightness_map[x+1][y]) +
				          (-1 * brightness_map[x-1][y+1]) + (1 * brightness_map[x+1][y+1]);
				y_sobel = (1 * brightness_map[x-1][y-1]) + (2 * brightness_map[x][y-1]) + (1 * brightness_map[x+1][y-1]) +
				          (-1 * brightness_map[x-1][y+1]) + (-2 * brightness_map[x][y+1]) + (-1 * brightness_map[x+1][y+1]);
				fragment_magnitude[x][y] = Math.sqrt(x_sobel * x_sobel + y_sobel * y_sobel);
				fragment_orientation[x][y] = Math.atan( (double) y_sobel / (double) x_sobel );
			}
		}


		// Consider all pixels with edge strength of 80 or higher as edge fragmenst and return the map of these
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (fragment_magnitude[x][y] > 80)
					edge_fragment_present[x][y] = true;
				else
					edge_fragment_present[x][y] = false;
			}
		}
		return edge_fragment_present;
	}


	/***************************************************************************/


	private void apply_mean_filter()
	// Applies a mean filter to the brightness data in brightness_map
	{
		int sum;
		double average;

		for (int y = 1; y < height - 1; y++)
		{
			for (int x = 1; x < width - 1; x++)
			{
				sum = brightness_map[x-1][y-1] + brightness_map[x-1][y] + brightness_map[x-1][y+1] +
				      brightness_map[x][y-1] + brightness_map[x][y] + brightness_map[x][y+1] +
				      brightness_map[x+1][y-1] + brightness_map[x+1][y] + brightness_map[x+1][y+1];
				brightness_map[x][y] = (int) ((double) sum / 9.0);
			}
		}
	}
}