/*
Cory McKay

ImagePanel Class

- This class holds the Panel that displays images in the Thresholder program.
  It also holds the methods which formulate and display the brightness histogram
  and convert the image to a binary image.
- The refreshImage method can be called to open and display a specified image
  file and calculate the brightness histogram for this image.
- The convertToBinaryImage method can be called to convert all pixels in the image
  with a brightness below the specified threshold to black, and all other pixels to white.
- The quadrantConvertToBinaryImage method can be called to calculate a different threshold
  for each of the four quadrants of the image and then convert the pixels of each quadrant
  using each quadrant's threshold so that all pixels in the image with a brightness below
  the specified threshold become black, and all other pixels become white.
- The drawHistogram method can be called to display the image with the brightness histogram
  superimposed over it. This also draws a vertical white line at the calculated threshold
  for the whole image.
- The smoothImage method can be called to apply a mean filter to the image. This helps to
  smooth out the image and remove noise.
- The expandWhite method can be called to turn all black pixels with a white neighbour to white.
- The expandBlack method can be called to turn all white pixels with a black neighbour to black.
- The automaticExpandWhite method can be called to automatically determine how many times white
  objects need to be expanded and then shrunk to remove black noise without distorting objects
  too much, and then do the expansion and shrinking.
- The automaticExpandBlack method can be called to automatically determine how many times black
  objects need to be expanded and then shrunk to remove white noise without distorting objects
  too much, and then do the expansion and shrinking.
- The getBackgroundColour method can be called to get the colour of the background of the image
  (black or white).
- The getWhiteExpansions method can be called to automatically determine how many times white
  objects need to be expanded and then shrunk to remove black noise without distorting objects
  too much.
- The getBlackExpansions method can be called to automatically determine how many times black
  objects need to be expanded and then shrunk to remove white noise without distorting objects
  too much.
- The processImage method can be called to process the image through the following operations:
	1) convert the image to a binary image
	2) apply a mean filter to the image
	3) determine the background colour of the image
	4) automatically determine how many expansions of the background colour need
	   to be made to remove noise without distorting the object too much
	5) expand and then shrink objects of the background colour the number of times
	   found in step 4
	6) repeat steps 4 and 5 with the foreground colour
	7) apply a mean filter to the image
- The findSeed method can be called to identify a seed region in the image image. This is the
  largest connected non-background region of a size greater than 30 pixels. This then colours
  the original image red in the area of this region. Returns a matrix with the indices corresponding
  to the pixels of the region set to true.
- The growRegion method can be called to identify a seed region in the image image using the findSeed
  method and then grow this region by adding connected pixels whose brightness is not too far outside
  the minimum and maximum intensities of the seed region. Keep adding pixels until no more can be added
  and then display the grown region in red on the original image.
- The detectEdgeFragments method can be called to identify all edges in the image in yellow.
- The detectRegionEdgesFirstStep, detectRegionEdgesSecondStep and detectRegionEdgesThirdStep
  methods can be called to find a region in the image, detect an edge around it and refine this edge
  so that it is one pixel thick and continuous.
*/

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.awt.image.*;
import java.io.File;

public class ImagePanel extends JPanel
{
	// Fields
	private BufferedImage image; // the image that is being displayed and processed
	private int width; // the widthh of the image
	private int height; // the height of the image
	private int hist[]; // the brightness level histogram for the image. The pixels are
	                    // given a brightness rating from 0 (dark) to 255 (light).
	                    // The historgram records the frequency of each of these brightness
	                    // levels in the image that is being processed.
	private boolean is_converted; // whether or not the image has been converted to binary
	private boolean histo_showing; // whether or not the brightness histogram is displayed
	private int[][] original_image_brightness_map; // a map of the unaltered image's brightness levels


	/***************************************************************************/


	public ImagePanel (String filename)
	// Constructor
	// Opens and displays the specified image file and creates a new histogram.
	// filename is the path of the image file to be processed.
	// NOTE: the program will automatically open cory.jpg if an invalid file is specified.
	{
		refreshImage(filename);
	}


	/***************************************************************************/


	public void refreshImage(String filename)
	// Opens the the specified image files and calculates the brightness histogram for it.
	// filename is the path of the image file to be processed.
	// Automatically open cory.jpg if an invalid file is specified.
	{
		int pixels[]; // the pixel information for the image


		// Load the image as an Image and then store it in the image BufferedImage field
		try
		{
			Image img = getToolkit().createImage(filename);
			MediaTracker t = new MediaTracker(this);
			t.addImage(img, 0);
			t.waitForAll();
			image = new BufferedImage(img.getWidth(this), img.getHeight(this),
				BufferedImage.TYPE_INT_RGB);
			image.createGraphics().drawImage(img, 0, 0, this);
		}
		catch (Exception e)
		{
			Image img = getToolkit().createImage("cory.jpg");
			MediaTracker t = new MediaTracker(this);
			t.addImage(img, 0);
			try
			{
				t.waitForAll();
			} catch (Exception f) {}
			image = new BufferedImage(img.getWidth(this), img.getHeight(this),
				BufferedImage.TYPE_INT_RGB);
			image.createGraphics().drawImage(img, 0, 0, this);
		}


		// Store the width and height of the object
		width = image.getWidth();
		height = image.getHeight();


		// Allow the surrounding ScrollPane to know when to use scrollbars
		setPreferredSize (new Dimension(width, height));


		// Reset the histogram and calculate its values for the entire given image
		hist = getQuadrantHistogram(0, width, 0, height);


		// Note that the image is not a binary image and the histogram is not showing
		is_converted = false;
		histo_showing = false;


		// Refresh the display of the image
		repaint();
	}


	/***************************************************************************/


	public void drawHistogram()
	// Display the image with the brightness histogram superimposed over it.
	// Also draws a vertical white line at the calculated threshold for the whole image.
	// Assumes that the histogram has already been calculated (i.e. refreshImage has been called).
	{
		// Calculate max_hist, the most common brightness range in the histogram
		int max_hist = 0;
		for (int i = 0; i < 256; i++)
		{
			if (hist[i] > max_hist)
				max_hist = hist[i];
		}


		// Draw the histogram
		Graphics2D g = image.createGraphics();
		int x; // the x-coordinate of a pixel
		int y; // the y-coordinate of a piexel
		int lasty = height - height * hist[0] / max_hist; // the last y coord that was drawn
		x = (width - 256) / 2;
		for (int i = 0; i < 256; i++, x++)
		{
			y = height - height * hist[i] / max_hist;
			g.setColor (new Color (i, i, i));
			g.fillRect (x, y, 1, height-y);
			g.setColor (Color.red);
			g.drawLine (x-1, lasty, x, y);
			lasty = y;
		}


		// Calculate the overall threshold of the image and draw a vertical white line
		// at that location on the histogram
		ThresholdCalculator calculator = new ThresholdCalculator(hist);
		int threshold = calculator.getThreshold();
		g.setColor (Color.white);
		g.drawLine ((width-256)/2 + threshold, 0, (width-256)/2 + threshold, height);


		/* FOR TESTING PURPOSES
		// Draw all significant local peaks as vertical blue lines
		int[] filtered_peaks = calculator.findAllPeaks();
		g.setColor (Color.blue);
		for (int i=0; i < filtered_peaks.length; i++)
		{
			g.drawLine ((width-256)/2 + filtered_peaks[i], 0, (width-256)/2 + filtered_peaks[i], height);
		}
		*/


		// Note that the image is not a binary image and the histogram is showing
		is_converted = false;
		histo_showing = true;


		// Refresh the display of the image
		repaint();
	}


	/***************************************************************************/


	public void convertToBinaryImage()
	// Convert all pixels in the image with a brightness below the calculated threshold
	// to black, and all other pixels to white.
	// Assumes that the histogram has already been calculated (i.e. refreshImage has been called).
	{
		int level; // the brightness value of each pixel
		int x; // the x-coordinate of a pixel
		int y; // the y-coordinate of a piexel
		int threshold; // the automatically determined threshold for the entire image


		// Calculate the overall threshold of the image
		ThresholdCalculator calculator = new ThresholdCalculator(hist);
		threshold = calculator.getThreshold();


		// Convert all pixels in the image to either black or white, based
		// on their brightness
		regionalConvertToBinary(0, width, 0, height, threshold);


		// Note that the image is now a binary image
		is_converted = true;


		// Refresh the display of the image
		repaint();
	}


	/***************************************************************************/


	public void quadrantConvertToBinaryImage()
	// Calculate a different threshold for each of the four quadrants of the image.
	// The four quadrants are top left, top right, bottom left and bottom right.
	// Convert the pixels of each quadrant using each quadrant's threshold so that
	// all pixels in the image with a brightness below the specified threshold
	// become black, and all other pixels become white.
	{
		ThresholdCalculator calculator; // used to calculate the quadrant thresholds
		int x_border = width / 2; // the x-coordinate where the quadrants split
		int y_border = height / 2; // the y-coordinate where the quadrants split


		// Get the histograms for each of the quadrants
		int[] top_left_hist = getQuadrantHistogram(0, x_border + 1, 0, y_border + 1);
		int[] top_right_hist = getQuadrantHistogram(x_border + 1, width, 0, y_border + 1);
		int[] bottom_left_hist = getQuadrantHistogram(0, x_border + 1, y_border + 1, height);
		int[] bottom_right_hist = getQuadrantHistogram(x_border + 1, width, y_border + 1, height);


		// Get the thresholds for each of the quadrants
		calculator = new ThresholdCalculator(top_left_hist);
		int top_left_threshold = calculator.getThreshold();
		calculator = new ThresholdCalculator(top_right_hist);
		int top_right_threshold = calculator.getThreshold();
		calculator = new ThresholdCalculator(bottom_left_hist);
		int bottom_left_threshold = calculator.getThreshold();
		calculator = new ThresholdCalculator(bottom_right_hist);
		int bottom_right_threshold = calculator.getThreshold();


		// Convert the quadrants of the image to binary
		regionalConvertToBinary(0, x_border + 1, 0, y_border + 1, top_left_threshold);
		regionalConvertToBinary(x_border + 1, width, 0, y_border + 1, top_right_threshold);
		regionalConvertToBinary(0, x_border + 1, y_border + 1, height, bottom_left_threshold);
		regionalConvertToBinary(x_border + 1, width, y_border + 1, height, bottom_right_threshold);


		// Note that the image is now a binary image
		is_converted = true;


		// Refresh the display of the image
		repaint();
	}


	/***************************************************************************/


	public void smoothImage(String filename)
	// Apply a mean filter to the image. This helps to smooth out the image and remove noise.
	// Convert the image to a binary image if it is not already one.
	{
		// Refresh the image and convert it to a binary image if necessary
		if (histo_showing == true)
			refreshImage(filename);
		if (is_converted == false)
			convertToBinaryImage();


		// Apply the mean filter
		NoiseRemover noise_remover = new NoiseRemover(image);
		image = noise_remover.getSmoothedImage();


		// Refresh the display of the image
		repaint();
	}


	/***************************************************************************/


	public void expandWhite(String filename)
	// Turn all black pixels with a white neighbour to white
	// Convert the image to a binary image if it is not already one.
	{
		// Refresh the image and convert it to a binary image if necessary
		if (histo_showing == true)
			refreshImage(filename);
		if (is_converted == false)
			convertToBinaryImage();


		// Expand white pixels
		NoiseRemover noise_remover = new NoiseRemover(image);
		image = noise_remover.whiteExpand();


		// Refresh the display of the image
		repaint();
	}


	/***************************************************************************/


	public void expandBlack(String filename)
	// Turn all white pixels with a black neighbour to black
	// Convert the image to a binary image if it is not already one.
	{
		// Refresh the image and convert it to a binary image if necessary
		if (histo_showing == true)
			refreshImage(filename);
		if (is_converted == false)
			convertToBinaryImage();


		// Expand black pixels
		NoiseRemover noise_remover = new NoiseRemover(image);
		image = noise_remover.blackExpand();


		// Refresh the display of the image
		repaint();
	}


	/***************************************************************************/


	public void automaticExpandWhite(String filename)
	// Automatically determine how many times white objects need to be expanded
	// and then shrunk to remove black noise without distorting objects too much,
	// and then do the expansion and shrinking.
	{
		// Refresh the image and convert it to a binary image if necessary
		if (histo_showing == true)
			refreshImage(filename);
		if (is_converted == false)
			convertToBinaryImage();


		// Expand and then shrink white objects
		NoiseRemover noise_remover = new NoiseRemover(image);
		int applications = noise_remover.getNumberWhiteExpansions();
		image = noise_remover.getWhiteExpandedShrunkImage(applications);


		// Refresh the display of the image
		repaint();
	}


	/***************************************************************************/


	public void automaticExpandBlack(String filename)
	// Automatically determine how many times black objects need to be expanded
	// and then shrunk to remove white noise without distorting objects too much,
	// and then do the expansion and shrinking.
	{
		// Refresh the image and convert it to a binary image if necessary
		if (histo_showing == true)
			refreshImage(filename);
		if (is_converted == false)
			convertToBinaryImage();


		// Expand and then shrink black objects
		NoiseRemover noise_remover = new NoiseRemover(image);
		int applications = noise_remover.getNumberBlackExpansions();
		image = noise_remover.getBlackExpandedShrunkImage(applications);


		// Refresh the display of the image
		repaint();
	}


	/***************************************************************************/


	public char getBackgroundColour(String filename)
	// Return the colour of the background of the image. This is determined by
	// counting the total number of white pixels and the total number of black pixels.
	// It is assumed that the colour with the larger number of pixels is the background
	// colour. Black is returned as 'b' and white is returned as 'w'.
	// Will convert the image to a binary image if it is not already.
	{
		// Refresh the image and convert it to a binary image if necessary
		if (histo_showing == true)
			refreshImage(filename);
		if (is_converted == false)
			convertToBinaryImage();


		// Determine the colour of the background
		NoiseRemover noise_remover = new NoiseRemover(image);
		return noise_remover.getColourOfBackground();
	}


	/***************************************************************************/


	public int getWhiteExpansions(String filename)
	// Automatically determine how many times white objects need to be expanded
	// and then shrunk to remove black noise without distorting objects too much.
	// Return this number.
	// Will convert the image to a binary image if it is not already.
	{
		// Refresh the image and convert it to a binary image if necessary
		if (histo_showing == true)
			refreshImage(filename);
		if (is_converted == false)
			convertToBinaryImage();


		// Calculate and return the number of expansions
		NoiseRemover noise_remover = new NoiseRemover(image);
		return noise_remover.getNumberWhiteExpansions();
	}


	/***************************************************************************/


	public int getBlackExpansions(String filename)
	// Automatically determine how many times black objects need to be expanded
	// and then shrunk to remove white noise without distorting objects too much.
	// Return this number.
	// Will convert the image to a binary image if it is not already.
	{
		// Refresh the image and convert it to a binary image if necessary
		if (histo_showing == true)
			refreshImage(filename);
		if (is_converted == false)
			convertToBinaryImage();


		// Calculate and return the number of expansions
		NoiseRemover noise_remover = new NoiseRemover(image);
		return noise_remover.getNumberBlackExpansions();
	}


	/***************************************************************************/


	public void processImage(String filename)
	// Process the image through the following operations:
	// 1) convert the image to a binary image
	// 2) apply a mean filter to the image
	// 3) determine the background colour of the image
	// 4) automatically determine how many expansions of the background colour need
	//    to be made to remove noise without distorting the object too much
	// 5) expand and then shrink objects of the background colour the number of times
	//    found in step 4
	// 6) repeat steps 4 and 5 with the foreground colour
	// 7) apply a mean filter to the image
	{
		// Refresh the image and convert it to a binary image
		refreshImage(filename);
		convertToBinaryImage();


		// Smooth the image using a mean filter
		smoothImage(filename);


		// Determine the colour of the background
		char background_colour = getBackgroundColour(filename);


		if (background_colour == 'w')
		{
			automaticExpandWhite(filename);
			automaticExpandBlack(filename);
		}
		else if (background_colour == 'b')
		{
			automaticExpandBlack(filename);
			automaticExpandWhite(filename);
		}


		// Smooth the image using a mean filter
		smoothImage(filename);
	}


	/***************************************************************************/


	public boolean[][] findSeed(String filename)
	// Identify a seed region in the image image. This is the largest connected non-background
	// region. Colours the original image red in the area of this region. Returns a matrix with
	// the indices corresponding to the pixels of the region set to true.
	{
		// Reload the image and convert it to a binary image
		refreshImage(filename);
		convertToBinaryImage();


		// Get the background and object colours of the image
		char background_colour = getBackgroundColour(filename);
		char object_colour;
		if (background_colour == 'w') object_colour = 'b';
		else object_colour = 'w';


		// Get a map of the image where the objects are labelled
		// and get a count of the number of pixels with each label
		char [][] pixels = new char[width][height];
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (findPixelBrightness(image.getRGB(x,y)) < 128)
					pixels[x][y] = 'b';
				else
					pixels[x][y] = 'w';
			}
		}
		ComponentLabeler labeler = new ComponentLabeler(pixels, width, height);
		MapCountObject map_count_object = labeler.getRegionMap(object_colour);


		// Restore the original image
		refreshImage(filename);


		// Find the largest labelled non-background object with size greater than 30 pixels
		int region_index = 0; // the index of the largest object
		int region_size = 0; // the number of pixels in the largest object
		for (int i = 1; i < map_count_object.number_pixels_per_object.length; i++) // i starts at 1 becasue must ignore background
		{
			if (map_count_object.number_pixels_per_object[i] > region_size)
			{
				region_index = i;
				region_size = map_count_object.number_pixels_per_object[i];
			}
		}


		// Create a map of the largest object, assigning true to those pixels belonging to the object
		boolean[][] map_of_an_object = new boolean[width][height]; // a map of an object
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (map_count_object.map[x][y] == region_index)
					map_of_an_object[x][y] = true;
				else map_of_an_object[x][y] = false;
			}
		}


		// Colour the detected object seed on the given object
		WritableRaster raster = image.getRaster();
		ColorModel model = image.getColorModel();
		Color label_colour = Color.red;
		int argbOC = label_colour.getRGB();
		Object colorData = model.getDataElements(argbOC, null);
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (map_of_an_object[x][y] == true)
					raster.setDataElements(x, y, colorData);
			}
		}

		// Return the map of the seed region in the image
		return map_of_an_object;
	}


	/***************************************************************************/


	public void growRegion(String filename)
	// Identify a seed region in the image image using the findSeed method.
	// Grow this region by adding connected pixels whose brightness is not too far
	// outside the minimum and maximum intensities of the seed region. Keep adding pixels
	// until no more can be added and then display the grown region in red on the original
	// image.
	{
		// Find the seed region and refresh the display
		boolean[][] map_of_an_object = findSeed(filename);
		refreshImage(filename);


		/* THE FOLLOWING IS CODE THAT CAN BE USED IN THE FUTURE FOR EDGE DETECTION
		   IT USES THE LAPLACIAN
		int edge_check;
		boolean change_made = true;
		while (change_made == true)
		{
			change_made = false;
			for (int y = 2; y < height - 2; y++)
			{
				for (int x = 2; x < width - 2; x++)
				{
					if (map_of_an_object[x][y] == true)
					{
						for (int i = x - 1; i <= x + 1; i++)
						{
							for (int j = y - 1; j <= y + 1; j++)
							{
								if (map_of_an_object[i][j] == false)
								{
									edge_check = findPixelBrightness(image.getRGB(i-1,j));
									edge_check = edge_check + findPixelBrightness(image.getRGB(i+1,j));
									edge_check = edge_check + findPixelBrightness(image.getRGB(i,j-1));
									edge_check = edge_check + findPixelBrightness(image.getRGB(i,j+1));
									edge_check = edge_check + (-4 * findPixelBrightness(image.getRGB(i,j)));

									if ( !(Math.abs(edge_check) >3) )
									{
										map_of_an_object[i][j] = true;
										change_made = true;
									}
								}
							}
						}
					}
				}
			}
		}
		*/


		// Determine the maximum and minimum intensities of the pixels in the region
		// as well as the average intensity
		int min_intensity = 256;
		int max_intensity = 1;
		double total = 0.0;
		int count = 0;
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (map_of_an_object[x][y] == true)
				{
					if (findPixelBrightness(image.getRGB(x,y)) > max_intensity)
						max_intensity = findPixelBrightness(image.getRGB(x,y));
					if (findPixelBrightness(image.getRGB(x,y)) < min_intensity)
						min_intensity = findPixelBrightness(image.getRGB(x,y));
					total = total + findPixelBrightness(image.getRGB(x,y));
					count++;
				}
			}
		}
		double average = total / (double) count;


		// Calculate the standard deviation of the intensities in the image
		double sum_so_far =0;
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (map_of_an_object[x][y] == true)
				{
					sum_so_far = sum_so_far +
						Math.pow((findPixelBrightness(image.getRGB(x,y)) - average), 2);
				}
			}
		}
		double standard_deviation = Math.sqrt(1.0/((double)(count-1))*sum_so_far);


		// Keep adding border pixels to the region that fall within an accepted range
		// of intensities (as defined by the souldAdd method) until no more
		// pixels are added.
		int intensity;
		boolean change_made = true;
		while (change_made == true)
		{
			change_made = false;
			for (int y = 1; y < height - 1; y++)
			{
				for (int x = 1; x < width - 1; x++)
				{
					if (map_of_an_object[x][y] == true)
					{
						for (int i = x - 1; i <= x + 1; i++)
						{
							for (int j = y - 1; j <= y + 1; j++)
							{
								if (map_of_an_object[i][j] == false)
								{
									intensity = findPixelBrightness(image.getRGB(i,j));
									if ( (shouldAdd(intensity, min_intensity, max_intensity, standard_deviation) == true) )
									{
										map_of_an_object[i][j] = true;
										change_made = true;
									}
								}
							}
						}
					}
				}
			}
		}


		// Colour the detected object seed on the given object
		WritableRaster raster = image.getRaster();
		ColorModel model = image.getColorModel();
		Color label_colour = Color.red;
		int argbOC = label_colour.getRGB();
		Object colorData = model.getDataElements(argbOC, null);
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (map_of_an_object[x][y] == true)
					raster.setDataElements(x, y, colorData);
			}
		}
	}


	/***************************************************************************/


	public void detectEdgeFragments()
	// Detect all edge fragments in the image and display them in orange on the image
	{
		EdgeDetector detector = new EdgeDetector(width, height, getBrightnessMap()); // used to detect edges
		boolean[][] map = detector.getFragments(true); // pixels where there is an edge are marked true


		// Colour the detected edges orange
		WritableRaster raster = image.getRaster();
		ColorModel model = image.getColorModel();
		Color edge_colour = Color.orange;
		int argbOC = edge_colour.getRGB();
		Object colorData = model.getDataElements(argbOC, null);
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (map[x][y] == true)
					raster.setDataElements(x, y, colorData);
			}
		}
	}


	/***************************************************************************/


	public int[][] detectRegionEdgesFirstStep(String filename)
	// Identify a region in the image image. Display it in red. Return a map
	// of the image where the region is bright and the rest of the image is not.
	// filename is the name of the image file.
	{
		int[][] brightness_map = new int[width][height]; //  a map of brightness levels


		// Reload the image and convert it to a binary image
		refreshImage(filename);
		convertToBinaryImage();


		// Get the background and object colours of the image
		char background_colour = getBackgroundColour(filename);
		char object_colour;
		if (background_colour == 'w') object_colour = 'b';
		else object_colour = 'w';


		// Get a map of the image where the objects are labelled
		// and get a count of the number of pixels with each label
		char [][] pixels = new char[width][height];
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (findPixelBrightness(image.getRGB(x,y)) < 128)
					pixels[x][y] = 'b';
				else
					pixels[x][y] = 'w';
			}
		}
		ComponentLabeler labeler = new ComponentLabeler(pixels, width, height);
		MapCountObject map_count_object = labeler.getRegionMap(object_colour);


		// Restore the original image
		// Store a map of image brightness in original_image_brightness_map
		refreshImage(filename);
		original_image_brightness_map = getBrightnessMap();


		// Find the largest labelled non-background region with size greater than 30 pixels
		int region_index = 0; // the index of the largest object
		int region_size = 0; // the number of pixels in the largest object
		for (int i = 1; i < map_count_object.number_pixels_per_object.length; i++) // i starts at 1 becasue must ignore background
		{
			if (map_count_object.number_pixels_per_object[i] > region_size)
			{
				region_index = i;
				region_size = map_count_object.number_pixels_per_object[i];
			}
		}


		// Create a map of the largest region
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (map_count_object.map[x][y] == region_index)
					brightness_map[x][y] = 255;
				else brightness_map[x][y] = 0;
			}
		}


		// Colour the detected region
		WritableRaster raster = image.getRaster();
		ColorModel model = image.getColorModel();
		Color label_colour = Color.red;
		int argbOC = label_colour.getRGB();
		Object colorData = model.getDataElements(argbOC, null);
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (brightness_map[x][y] == 255)
					raster.setDataElements(x, y, colorData);
			}
		}
		repaint();


		// Return the map of the region
		return brightness_map;
	}


	/***************************************************************************/


	public boolean[][] detectRegionEdgesSecondStep(int[][] brightness_map)
	// Detect edge fragments in the region defined on brightness_map and display them in orange on the image.
	// Return a map of the edges, with the coordinates where an edge is detected marked true.
	{
		EdgeDetector detector = new EdgeDetector(width, height, brightness_map); // used to detect edges
		boolean[][] map = detector.getFragments(false); // pixels where there is an edge are marked true


		// Colour the detected edges orange
		WritableRaster raster = image.getRaster();
		ColorModel model = image.getColorModel();
		Color edge_colour = Color.orange;
		int argbOC = edge_colour.getRGB();
		Object colorData = model.getDataElements(argbOC, null);
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (map[x][y] == true)
					raster.setDataElements(x, y, colorData);
			}
		}
		repaint();


		// Return the map
		return map;
	}


	/***************************************************************************/


	public void detectRegionEdgesThirdStep(boolean[][] edge_map)
	// Refine the edges on edge_map by following the boundary following
	{
		boolean[][] refined_boundary = new boolean[width][height]; // a map of the refined boundary


		// Set the extreme borders of edge_map to false so that the boundary following algorithm
		// will function properly. Also initialize refined_boundary.
		for (int i = 0; i < width; i++)
		{
			edge_map[i][0] = false;
			edge_map[i][height - 1] = false;
		}
		for (int i = 0; i < height; i++)
		{
			edge_map[0][i] = false;
			edge_map[width - 1][i] = false;
		}
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				refined_boundary[x][y] = false;
			}
		}


		// Find the first pixel of the edge through a left to right top to bottom scan
		int starting_x = 0; // the x-coordinate of the first pixel of the boundary that is detected
		int starting_y = 0; // the y-coordinate of the first pixel of the boundary that is detected
		EdgeDetector detector = new EdgeDetector(width, height, original_image_brightness_map); // used to detect edges
		for (int y = 1; y < height - 1; y++)
		{
			for (int x = 1; x < width - 1; x++)
			{
				if (edge_map[x][y] == true)
				{
					starting_x = x;
					starting_y = y;
					x = width;
					y = height;
				}
			}
		}


		// Use boundary following to get a thinned edge of the region. If there are any gaps in the
		// image, fill them in by selecting the pixels in the 8-neighbourhood of c with the strongest edge
		// strengths (in the original image) that are not already selected as part of the final edge.
		BoundaryObject next_b;
		int cx = starting_x;
		int cy = starting_y;
		int bx = cx - 1;
		int by = cy;
		int old_bx = bx;
		int old_by = by;
		int j;
		do // loop until returned to the starting position
		{
			refined_boundary[cx][cy] = true; // include the current c point in the final edge
			for (j = 1; j <= 8; j++) // basic boundary following algorithm
			{
				if (edge_map[bx][by] == true)
				{
					cx = bx;
					cy = by;
					bx = old_bx;
					by = old_by;
					j = 10;
				}
				else
				{
					next_b = moveClockwise(bx, by, cx, cy);
					old_bx = bx;
					old_by = by;					
					bx = next_b.x;
					by = next_b.y;
				}
			}
		} while (cx != starting_x || cy != starting_y);


		// Colour the final detected edge green
		WritableRaster raster = image.getRaster();
		ColorModel model = image.getColorModel();
		Color edge_colour = Color.green;
		int argbOC = edge_colour.getRGB();
		Object colorData = model.getDataElements(argbOC, null);
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (refined_boundary[x][y] == true)
					raster.setDataElements(x, y, colorData);
			}
		}
		repaint();
	}


	/***************************************************************************/


	public void paintComponent (Graphics g)
	// Allow the image and changes to the image to be displayed
	{
		super.paintComponent (g);
		g.drawImage (image, 0, 0, null);
	}


	/***************************************************************************/


	private int findPixelBrightness(int pixel_data)
	// Returns the brightness of the pixel given in pixel_data
	{
		int red   = (0x00ff0000 & pixel_data) >> 16;
		int green = (0x0000ff00 & pixel_data) >> 8;
		int blue  = (0x000000ff & pixel_data);
		/* return (int)(.33*red + .56*green + .11*blue); THIS IS HUMAN BALANCED */
		return (int)(.333*red + .333*green + .333*blue);
	}


	/***************************************************************************/


	private int[] getQuadrantHistogram(int x_left, int x_right, int y_top, int y_bottom)
	// Returns a brightness histogram of image within the given boundary coordinates
	{
		int[] quadrant_hist = new int[256]; // the histogram for the given area of the image


		// Set all values of quadrant_hist to 0
		for (int i = 0; i < 256; i++)
		{
			quadrant_hist[i] = 0;
		}


		// Fill up and return the histogram for the given area of the image
		for (int y = y_top; y < y_bottom; y++)
		{
			for (int x = x_left; x < x_right; x++)
			{
				quadrant_hist[ findPixelBrightness( image.getRGB(x,y) ) ]++;
			}
		}
		return quadrant_hist;
	}


	/***************************************************************************/


	private int[][] getBrightnessMap()
	// Returns a map of the image's brightness levels, from 0 to 255
	{
		int[][] map = new int[width][height]; // the map of the image's brightness levels


		// Plot the brightness levels on the map
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				map[x][y] = findPixelBrightness( image.getRGB(x,y) );
			}
		}
		return map;
	}


	/***************************************************************************/


	private void regionalConvertToBinary(int x_left, int x_right, int y_top, int y_bottom,
	                                     int threshold)
	// Convert all pixels in withen the given boundaries of the image with a brightness below
	// the specified threshold to black, and all other pixels to white.
	{
		WritableRaster raster = image.getRaster();
		ColorModel model = image.getColorModel();
		Color white = Color.white;
		Color black = Color.black;
		int argbWhite = white.getRGB();
		int argbBlack = black.getRGB();
		Object colorDataWhite = model.getDataElements(argbWhite, null);
		Object colorDataBlack = model.getDataElements(argbBlack, null);


		// Convert those pixels within the given boundaries to either black or white, based
		// on their brightness
		for (int y = y_top; y < y_bottom; y++)
		{
			for (int x= x_left; x < x_right; x++)
			{
				if (findPixelBrightness(image.getRGB(x,y)) < threshold)
					raster.setDataElements(x, y, colorDataBlack);
				else
					raster.setDataElements(x, y, colorDataWhite);
			}
		}
	}


	/***************************************************************************/


	private boolean shouldAdd(int intensity, int min, int max, double standard_deviation)
	// Return true if a given intensity is between window units less than the min
	// intensity and window units more than the max intensity. Window is defined as
	// twice the given standard deviation, with a ceiling at 15.
	{
		double window = 2 * standard_deviation;
		if (window > 15) standard_deviation = 15;
		if ( (intensity >= min - standard_deviation) && (intensity <= max + standard_deviation) ) return true;
		else return false;
	}


	/***************************************************************************/


	private BoundaryObject moveClockwise(int bx, int by, int cx, int cy)
	// Takes in the coordinates of a center point (cx, xy) and one of its 8-neighbours (bx, by) and
	// returns the next 8-neighbour of in the clockwise direction.
	{
		if (bx == cx - 1 && by == cy - 1)
			return new BoundaryObject(cx, cy - 1);
		else if (bx == cx && by == cy - 1)
			return new BoundaryObject(cx + 1, cy - 1);
		else if (bx == cx + 1 && by == cy - 1)
			return new BoundaryObject(cx + 1, cy);
		else if (bx == cx + 1 && by == cy)
			return new BoundaryObject(cx + 1, cy + 1);
		else if (bx == cx + 1 && by == cy + 1)
			return new BoundaryObject(cx, cy + 1);
		else if (bx == cx && by == cy + 1)
			return new BoundaryObject(cx - 1, cy + 1);
		else if (bx == cx - 1 && by == cy + 1)
			return new BoundaryObject(cx - 1, cy);
		else // if (bx == cx - 1 && by == cy)
			return new BoundaryObject(cx - 1, cy - 1 );
	}
}