/*
Cory McKay

MainFrame Class

- This class contains the code for the main part of the GUI of the program.
- An ImagePanel object is created and displayed in a ScrollPane.
- There is a set of menus that allow the user to perform all of the functions of this
  program. They can be used to load image files, view the brightness histogram
  for these files, convert the image into a binary image with an appropriate automatically
  determined threshold to bring out an object and break the image into 4 parts and apply the
  thresholding algorithm separately to each part. They can also be used to remove noise from
  binary images by using a mean filter and by expanding and shrinking objects. The user can
  have this all done automatically, or can control the intermediate steps. The program can
  explain to the user what is being done. The program can also detect a seed region and grow it.
  The program cans also detect edges in the image as well as edges around a region in an image
- There is a panel of buttons which allows the user to perform the most common tasks. These
  are loading a new image, refreshing an image that has been altered to the version on disk
  and performing a transformation to a binary image with noise removal, along with a readout
  of what was done.
- The Demonstrate Processing command begins a step by step tour of the image processing process.
  Most interface options are disabled during the tour to prevent the user from interfering with
  the proper execution of the image processing process. This can be cancelled at any time by pressing the
  cancel button. The next step of the tour is accessed by pressing the Next Step button.
- There is an area where images are displayed.
- There is an output area on the bottom of the screen where information is printed out to the
  user.
- This frame is resizable.
- NOTE: the program will automatically open cory.jpg if an invalid file is specified.
*/


import java.io.File;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.awt.image.*;

class MainFrame extends JFrame
{
	// Non-GUI fields
	private String imagename; // the path of the current file


	// General GUI fields
	private Container contentPane; // hold all elements of the GUI
	private ImagePanel thresholder; // the object used to store, display and process each image
	private JScrollPane scrollpane; // the ScrollPane used to store the ImagePanel
	private JTextArea textArea; // the area used to give the user feedback
	private JScrollPane scrollpane2; // the ScrollPane used to store textArea
	private JFileChooser chooser; // a dialog box to open files


	// Menu fields
	private JMenuBar menu_bar;
	private JMenu file_menu;
	private JMenu conversion_menu;
	private JMenu noise_menu;
	private JMenu growing_menu;
	private JMenu edge_menu;
	private JMenuItem open_item;
	private JMenuItem exit_item;
	private JMenuItem restore_item;
	private JMenuItem show_histogram_item;
	private JMenuItem quadrant_conversion_item;
	private JMenuItem overall_conversion_item;
	private JMenuItem smooth_item;
	private JMenuItem expand_white_item;
	private JMenuItem expand_black_item;
	private JMenuItem get_background_colour_item;
	private JMenuItem get_white_expansions_item;
	private JMenuItem get_black_expansions_item;
	private JMenuItem automatic_expand_white_item;
	private JMenuItem automatic_expand_black_item;
	private JMenuItem process_image_item;
	private JMenuItem demonstrate_item;
	private JMenuItem find_seed_regions_item;
	private JMenuItem grow_seed_region_item;
	private JMenuItem detect_edges_item;
	private JMenuItem detect_region_edges_item;


	// Button fields
	private JPanel buttonPanel;
	private JButton loadButton;
	private JButton restoreButton;
	private JButton demonstrateButton;
	private JButton edgeButton;


	// Demonstration fields
	private int demo_state; // how many steps have been completed in the demonstration so far
	private char background_colour; // the colour that has been determined as the background
	private int expansions; // the number of expansions that are determined to be needed
	private int[][] object_map; // used by DetectRegionEdgeFragmentsAction
	private boolean[][] edge_map; // used by DetectRegionEdgeFragmentsAction


	/***************************************************************************/


	public MainFrame(String filename)
	// Constructor
	// The filename is either the path of an image file or null
	{
		// Configure the frame
		setTitle("ImageProcessor");
		chooser = null;
		demo_state = 0;


		// Cause the program to quit when the exit box is pressed
		addWindowListener(new WindowAdapter() {
			public void windowClosing (WindowEvent e) {
				System.exit(0);
			}
		});


		// Load an image and display it in the thresholder ImagePanel in a ScrolPane
		// If a filename was specified by the calling object, then this file is opened.
		// If a filename was not specified, then a JFileChooser dialog box is shown on startup.
		// NOTE: the program will automatically open cory.jpg if an invalid file is specified.
		if (filename != null)
			imagename = filename;
		else
		{
			int chooser_return;
			chooser = new JFileChooser();
			chooser.setCurrentDirectory(new File("."));
			chooser_return = chooser.showOpenDialog(MainFrame.this);
			if (chooser_return != JFileChooser.APPROVE_OPTION)
				System.exit(0);
			imagename = chooser.getSelectedFile().getPath();
		}
		thresholder = new ImagePanel(imagename);
		scrollpane = new JScrollPane();
		scrollpane.setViewportView(thresholder);


		// Display the text area where feedback to the user is given. Put it in a ScrollPane
		textArea = new JTextArea();
		textArea.setLineWrap(false);
		textArea.setRows(6);
		textArea.setColumns(45);
		textArea.setEditable(false);
		scrollpane2 = new JScrollPane(textArea);
		textArea.setText("");


		// Set up the File menu
		file_menu = new JMenu("File");
		open_item = new JMenuItem("Load Image");
		file_menu.add(open_item);
		open_item.addActionListener(new OpenAction());
		restore_item = new JMenuItem("Restore Image");
		file_menu.add(restore_item);
		restore_item.addActionListener(new RestoreAction());
		file_menu.addSeparator();
		exit_item = new JMenuItem("Exit");
		file_menu.add(exit_item);
		exit_item.addActionListener(new ExitAction());


		// Set up the Binary Conversion menu
		conversion_menu = new JMenu("Binary Conversion");
		show_histogram_item = new JMenuItem("Show Histogram");
		conversion_menu.add(show_histogram_item);
		show_histogram_item.addActionListener(new HistoAction());
		conversion_menu.addSeparator();
		quadrant_conversion_item = new JMenuItem("Quadrant Conversion");
		conversion_menu.add(quadrant_conversion_item);
		quadrant_conversion_item.addActionListener(new QuadrantConversionAction());
		overall_conversion_item = new JMenuItem("Overall Conversion");
		conversion_menu.add(overall_conversion_item);
		overall_conversion_item.addActionListener(new OverallConversionAction());


		// Set up the Noise Removal menu
		noise_menu = new JMenu("Noise Removal");
		smooth_item = new JMenuItem("Apply Mean Filter");
		noise_menu.add(smooth_item);
		smooth_item.addActionListener(new SmoothAction());
		noise_menu.addSeparator();
		expand_white_item = new JMenuItem("Expand White / Shrink Black");
		noise_menu.add(expand_white_item);
		expand_white_item.addActionListener(new ExpandWhiteAction());
		expand_black_item = new JMenuItem("Expand Black / Shrink White");
		noise_menu.add(expand_black_item);
		expand_black_item.addActionListener(new ExpandBlackAction());
		noise_menu.addSeparator();
		get_background_colour_item = new JMenuItem("Get Background Colour");
		noise_menu.add(get_background_colour_item);
		get_background_colour_item.addActionListener(new GetBackgroundColourAction());
		get_white_expansions_item = new JMenuItem("Calculate White Expansions");
		noise_menu.add(get_white_expansions_item);
		get_white_expansions_item.addActionListener(new GetWhiteExpansionsAction());
		get_black_expansions_item = new JMenuItem("Calculate Black Expansions");
		noise_menu.add(get_black_expansions_item);
		get_black_expansions_item.addActionListener(new GetBlackExpansionsAction());
		noise_menu.addSeparator();
		automatic_expand_white_item = new JMenuItem("Automatic Expand Black");
		noise_menu.add(automatic_expand_white_item);
		automatic_expand_white_item.addActionListener(new AutomaticExpandWhiteAction());
		automatic_expand_black_item = new JMenuItem("Automatic Expand White");
		noise_menu.add(automatic_expand_black_item);
		automatic_expand_black_item.addActionListener(new AutomaticExpandBlackAction());
		noise_menu.addSeparator();
		process_image_item = new JMenuItem("Process Image");
		noise_menu.add(process_image_item);
		process_image_item.addActionListener(new ProcessImageAction());
		demonstrate_item = new JMenuItem("Noise Removal Tour");
		noise_menu.add(demonstrate_item);
		demonstrate_item.addActionListener(new DemonstrateAction());


		// Set up the Region Growing menu
		growing_menu = new JMenu("Region Growing");
		find_seed_regions_item = new JMenuItem("Show Seed Region");
		growing_menu.add(find_seed_regions_item);
		find_seed_regions_item.addActionListener(new FindSeedAction());
		grow_seed_region_item = new JMenuItem("Grow A Region");
		growing_menu.add(grow_seed_region_item);
		grow_seed_region_item.addActionListener(new GrowSeedAction());


		// Set up the Edge Detection menu
		edge_menu = new JMenu("Edge Detection");
		detect_edges_item = new JMenuItem("Detect All Edges");
		edge_menu.add(detect_edges_item);
		detect_edges_item.addActionListener(new DetectEdgeFragmentsAction());
		detect_region_edges_item = new JMenuItem("Region Outlining Tour");
		edge_menu.add(detect_region_edges_item);
		detect_region_edges_item.addActionListener(new DetectRegionEdgeFragmentsAction());


		// Set up the menu bar
		menu_bar = new JMenuBar();
		setJMenuBar(menu_bar);
		menu_bar.add(file_menu);
		menu_bar.add(conversion_menu);
		menu_bar.add(noise_menu);
		menu_bar.add(growing_menu);
		menu_bar.add(edge_menu);


		// Set up the buttons
		loadButton = new JButton("Load Image");
		loadButton.addActionListener(new OpenAction());
		restoreButton = new JButton("Restore Image");
		restoreButton.addActionListener(new RestoreAction());
		demonstrateButton = new JButton("Noise Removal Tour");
		demonstrateButton.addActionListener(new DemonstrateAction());
		edgeButton = new JButton("Region Outlining Tour");
		edgeButton.addActionListener(new DetectRegionEdgeFragmentsAction());


		// Set up the panel of buttons
		buttonPanel = new JPanel();
		buttonPanel.add(loadButton);
		buttonPanel.add(restoreButton);
		buttonPanel.add(demonstrateButton);
		buttonPanel.add(edgeButton);


		// Display the ImagePanel in the centre of the Frame, the buttonPanel
		// at the top of the Frame and the textArea at the bottom of the Frame
		contentPane = getContentPane();
		contentPane.add(buttonPanel, BorderLayout.NORTH);
		contentPane.add(scrollpane, BorderLayout.CENTER);
		contentPane.add(scrollpane2, BorderLayout.SOUTH);
	}


	/***************************************************************************/


	private void demonstrateBegin()
	// Reset the interface for when the binary conversion tour is started. This is to prevent the user
	// from taking actions which might interfere with the correct operation of the tour.
	{
		// Reset the command names
		demonstrateButton.setText("Next Step");
		demonstrate_item.setText("Next Step");
		restore_item.setText("Cancel Demo");
		restoreButton.setText("Cancel Demo");


		// Disable options
		loadButton.setEnabled(false);
		edgeButton.setEnabled(false);
		open_item.setEnabled(false);
		show_histogram_item.setEnabled(false);
		quadrant_conversion_item.setEnabled(false);
		overall_conversion_item.setEnabled(false);
		smooth_item.setEnabled(false);
		expand_white_item.setEnabled(false);
		expand_black_item.setEnabled(false);
		get_background_colour_item.setEnabled(false);
		get_white_expansions_item.setEnabled(false);
		get_black_expansions_item.setEnabled(false);
		automatic_expand_white_item.setEnabled(false);
		automatic_expand_black_item.setEnabled(false);
		process_image_item.setEnabled(false);
		find_seed_regions_item.setEnabled(false);
		grow_seed_region_item.setEnabled(false);
		detect_edges_item.setEnabled(false);
		detect_region_edges_item.setEnabled(false);
	}


	/***************************************************************************/


	private void demonstrateEnd()
	// Reset the interface for when the tour is ended.
	{
		// Reset the command names
		demo_state = 0;
		demonstrateButton.setText("Noise Removal Tour");
		demonstrate_item.setText("Noise Removal Tour");
		restore_item.setText("Restore Image");
		restoreButton.setText("Restore Image");
		edgeButton.setText("Region Outlining Tour");
		detect_region_edges_item.setText("Region Outlining Tour");


		// Enable options
		loadButton.setEnabled(true);
		demonstrateButton.setEnabled(true);
		edgeButton.setEnabled(true);
		open_item.setEnabled(true);
		show_histogram_item.setEnabled(true);
		quadrant_conversion_item.setEnabled(true);
		overall_conversion_item.setEnabled(true);
		smooth_item.setEnabled(true);
		expand_white_item.setEnabled(true);
		expand_black_item.setEnabled(true);
		get_background_colour_item.setEnabled(true);
		get_white_expansions_item.setEnabled(true);
		get_black_expansions_item.setEnabled(true);
		automatic_expand_white_item.setEnabled(true);
		automatic_expand_black_item.setEnabled(true);
		process_image_item.setEnabled(true);
		demonstrate_item.setEnabled(true);
		find_seed_regions_item.setEnabled(true);
		grow_seed_region_item.setEnabled(true);
		detect_edges_item.setEnabled(true);
		detect_region_edges_item.setEnabled(true);
	}


	/***************************************************************************/


	private void demonstrateEdgeDetectionBegin()
	// Reset the interface for when the edge detection tour is started. This is to prevent the user
	// from taking actions which might interfere with the correct operation of the tour.
	{
		// Reset the command names
		edgeButton.setText("Next Step");
		detect_region_edges_item.setText("Next Step");
		restore_item.setText("Cancel Demo");
		restoreButton.setText("Cancel Demo");


		// Disable options
		loadButton.setEnabled(false);
		demonstrateButton.setEnabled(false);
		open_item.setEnabled(false);
		show_histogram_item.setEnabled(false);
		quadrant_conversion_item.setEnabled(false);
		overall_conversion_item.setEnabled(false);
		smooth_item.setEnabled(false);
		expand_white_item.setEnabled(false);
		expand_black_item.setEnabled(false);
		get_background_colour_item.setEnabled(false);
		get_white_expansions_item.setEnabled(false);
		get_black_expansions_item.setEnabled(false);
		automatic_expand_white_item.setEnabled(false);
		automatic_expand_black_item.setEnabled(false);
		process_image_item.setEnabled(false);
		demonstrate_item.setEnabled(false);
		find_seed_regions_item.setEnabled(false);
		grow_seed_region_item.setEnabled(false);
		detect_edges_item.setEnabled(false);
	}


	/***************************************************************************/


	class OpenAction extends AbstractAction
	// Bring up a JFileChooser dialog box to open new image files
	// NOTE: will automatically open cory.jpg if an invalid file is specified.
	{
		public void actionPerformed(ActionEvent event)
		{
			int chooser_return;
			textArea.setText("");
			if (chooser == null) // create a new JFileChooser if one doesn't exist already
			{
				chooser = new JFileChooser();
				chooser.setCurrentDirectory(new File("."));
			}
			chooser_return = chooser.showOpenDialog(MainFrame.this);
			if (chooser_return == JFileChooser.APPROVE_OPTION) // only do if OK chosen
			{
				imagename = chooser.getSelectedFile().getPath();
				thresholder.refreshImage(imagename);
				MainFrame.this.scrollpane.revalidate();
			}
		}
	}


	/***************************************************************************/


	class RestoreAction extends AbstractAction
	// Restore the unaltered image that is currently loaded
	{
		public void actionPerformed(ActionEvent event)
		{
			demonstrateEnd();
			textArea.setText("");
			thresholder.refreshImage(imagename);
		}
	}


	/***************************************************************************/


	class ExitAction extends AbstractAction
	// Quit the program after asking the user if he wishes to save his work
	{
		public void actionPerformed(ActionEvent event)
		{
				System.exit(0);
		}
	}


	/***************************************************************************/


	class HistoAction extends AbstractAction
	// Display the brightness histogram of the current image along with the
	// calculated threshold.
	{
		public void actionPerformed(ActionEvent event)
		{
			textArea.setText("");
			thresholder.refreshImage(imagename);
			thresholder.drawHistogram();
		}
	}


	/***************************************************************************/


	class QuadrantConversionAction extends AbstractAction
	// Convert the current image to a binary image, with a different threshold that is
	// automatically determined individually for each of the four quadrants of the image.
	{
		public void actionPerformed(ActionEvent event)
		{
			textArea.setText("");
			thresholder.refreshImage(imagename);
			thresholder.quadrantConvertToBinaryImage();
		}
	}


	/***************************************************************************/


	class OverallConversionAction extends AbstractAction
	// Convert the current image to a binary image, with a threshold that is automatically
	// determined to bring out an object in the image.
	{
		public void actionPerformed(ActionEvent event)
		{
			textArea.setText("");
			thresholder.refreshImage(imagename);
			thresholder.convertToBinaryImage();
		}
	}


	/***************************************************************************/


	class SmoothAction extends AbstractAction
	// Apply a mean filter to the image. This helps to smooth out the image and remove noise.
	{
		public void actionPerformed(ActionEvent event)
		{
			textArea.setText("");
			thresholder.smoothImage(imagename);
		}
	}


	/***************************************************************************/


	class ExpandWhiteAction extends AbstractAction
	// Turn all black pixels with a white neighbour to white
	{
		public void actionPerformed(ActionEvent event)
		{
			textArea.setText("");
			thresholder.expandWhite(imagename);
		}
	}


	/***************************************************************************/


	class ExpandBlackAction extends AbstractAction
	// Turn all white pixels with a black neighbour to black
	{
		public void actionPerformed(ActionEvent event)
		{
			textArea.setText("");
			thresholder.expandBlack(imagename);
		}
	}


	/***************************************************************************/


	class GetBackgroundColourAction extends AbstractAction
	// Give the user the colour of the background
	{
		public void actionPerformed(ActionEvent event)
		{
			char colour = thresholder.getBackgroundColour(imagename);
			if (colour == 'w')
				textArea.setText("The background colour is white.");
			else if (colour == 'b')
				textArea.setText("The background colour is black.");
		}
	}


	/***************************************************************************/


	class GetWhiteExpansionsAction extends AbstractAction
	// Automatically determine how many times white objects need to be expanded
	// and then shrunk to remove black noise without distorting objects too much.
	// Print out this number.
	// Will convert the image to a binary image if it is not already.
	{
		public void actionPerformed(ActionEvent event)
		{
			int number_expansions = thresholder.getWhiteExpansions(imagename);
			textArea.setText("White objects should be expanded and then shrunk " +
			                 number_expansions + " times.");
		}
	}


	/***************************************************************************/


	class GetBlackExpansionsAction extends AbstractAction
	// Automatically determine how many times black objects need to be expanded
	// and then shrunk to remove white noise without distorting objects too much.
	// Print out this number.
	// Will convert the image to a binary image if it is not already.
	{
		public void actionPerformed(ActionEvent event)
		{
			int number_expansions = thresholder.getBlackExpansions(imagename);
			textArea.setText("Black objects should be expanded and then shrunk " +
			                 number_expansions + " times.");
		}
	}


	/***************************************************************************/


	class AutomaticExpandWhiteAction extends AbstractAction
	// Automatically determine how many times white objects need to be expanded
	// and then shrunk to remove black noise without distorting objects too much,
	// and then do the expansion and shrinking.
	{
		public void actionPerformed(ActionEvent event)
		{
			textArea.setText("");
			thresholder.automaticExpandWhite(imagename);
		}
	}


	/***************************************************************************/


	class AutomaticExpandBlackAction extends AbstractAction
	// Automatically determine how many times black objects need to be expanded
	// and then shrunk to remove white noise without distorting objects too much
	// and then do the expansion and shrinking.
	{
		public void actionPerformed(ActionEvent event)
		{
			textArea.setText("");
			thresholder.automaticExpandBlack(imagename);
		}
	}


	/***************************************************************************/


	class ProcessImageAction extends AbstractAction
	// Process the image through the following operations:
	// 1) convert the image to a binary image
	// 2) apply a mean filter to the image
	// 3) determine the background colour of the image
	// 4) automatically determine how many expansions of the background colour need
	//    to be made to remove noise without distorting the object too much
	// 5) expand and then shrink objects of the background colour the number of times
	//    found in step 4
	// 6) repeat steps 4 and 5 with the foreground colour
	// 7) apply a mean filter to the image
	{
		public void actionPerformed(ActionEvent event)
		{
			textArea.setText("");
			thresholder.processImage(imagename);
		}
	}


	/***************************************************************************/


	class DemonstrateAction extends AbstractAction
	// Process the image and give information explaining how this
	// is being done. The image is processed through the following operations:
	// 1) convert the image to a binary image
	// 2) apply a mean filter to the image
	// 3) determine the background colour of the image
	// 4) automatically determine how many expansions of the background colour need
	//    to be made to remove noise without distorting the object too much
	// 5) expand and then shrink objects of the background colour the number of times
	//    found in step 4
	// 6) repeat steps 4 and 5 with the foreground colour
	// 7) apply a mean filter to the image
	{
		public void actionPerformed(ActionEvent event)
		{
			if (demo_state == 0)
			{
				textArea.setText("PROCESSING " + imagename + ":\n");
				textArea.append("- Reloading the image... ");
				thresholder.refreshImage(imagename);
				textArea.append(" image succesfully reloaded.\n");
				demonstrateBegin();
			}
			if (demo_state == 1)
			{
				textArea.append("- Converting to binary image... ");
				thresholder.convertToBinaryImage();
				textArea.append(" conversion succesful.\n");
			}
			if (demo_state == 2)
			{
				textArea.append("- Applying mean filter... ");
				thresholder.smoothImage(imagename);
				textArea.append(" filter succesfully applied.\n");
			}
			if (demo_state == 3)
			{
				textArea.append("- Determining background colour... ");
				background_colour = thresholder.getBackgroundColour(imagename);
				if (background_colour == 'w')
					textArea.append(" background is white.\n");
				else if (background_colour == 'b')
					textArea.append(" background is black.\n");
			}
			if (demo_state == 4)
			{
				if (background_colour == 'w')
				{
					textArea.append("- Determining times to expand/shrink white... ");
					expansions = thresholder.getWhiteExpansions(imagename);
					textArea.append(expansions + " expansions recommended.\n");
				}
				else if (background_colour == 'b')
				{
					textArea.append("- Determining times to expand/shrink black... ");
					expansions = thresholder.getBlackExpansions(imagename);
					textArea.append(expansions + " expansions recommended.\n");
				}
			}
			if (demo_state == 5)
			{
				if (background_colour == 'w')
				{
					textArea.append("- Expanding white... ");
					for (int i = 0; i < expansions; i++)
						thresholder.expandWhite(imagename);
					textArea.append("successfully expanded white " + expansions + " times.\n");
				}
				else if (background_colour == 'b')
				{
					textArea.append("- Expanding black... ");
					for (int i = 0; i < expansions; i++)
						thresholder.expandBlack(imagename);
					textArea.append("successfully expanded black " + expansions + " times.\n");
				}
			}
			if (demo_state == 6)
			{
				if (background_colour == 'w')
				{
					textArea.append("- Shrinking white... ");
					for (int i = 0; i < expansions; i++)
						thresholder.expandBlack(imagename);
					textArea.append("successfully shrunk white " + expansions + " times.\n");
				}
				else if (background_colour == 'b')
				{
					textArea.append("- Shrinking black... ");
					for (int i = 0; i < expansions; i++)
						thresholder.expandWhite(imagename);
					textArea.append("successfully shrunk black " + expansions + " times.\n");
				}
			}
			if (demo_state == 7)
			{
				if (background_colour == 'w')
				{
					textArea.append("- Determining times to expand/shrink black... ");
					expansions = thresholder.getBlackExpansions(imagename);
					textArea.append(expansions + " expansions recommended.\n");
				}
				else if (background_colour == 'b')
				{
					textArea.append("- Determining times to expand/shrink white... ");
					expansions = thresholder.getWhiteExpansions(imagename);
					textArea.append(expansions + " expansions recommended.\n");
				}
			}
			if (demo_state == 8)
			{
				if (background_colour == 'w')
				{
					textArea.append("- Expanding black... ");
					for (int i = 0; i < expansions; i++)
						thresholder.expandBlack(imagename);
					textArea.append("successfully expanded black " + expansions + " times.\n");
				}
				else if (background_colour == 'b')
				{
					textArea.append("- Expanding white... ");
					for (int i = 0; i < expansions; i++)
						thresholder.expandWhite(imagename);
					textArea.append("successfully expanded white " + expansions + " times.\n");
				}
			}
			if (demo_state == 9)
			{
				if (background_colour == 'w')
				{
					textArea.append("- Shrinking black... ");
					for (int i = 0; i < expansions; i++)
						thresholder.expandWhite(imagename);
					textArea.append("successfully shrunk black " + expansions + " times.\n");
				}
				else if (background_colour == 'b')
				{
					textArea.append("- Shrinking white... ");
					for (int i = 0; i < expansions; i++)
						thresholder.expandBlack(imagename);
					textArea.append("successfully shrunk white " + expansions + " times.\n");
				}
			}
			if (demo_state == 10)
			{
				textArea.append("- Applying mean filter... ");
				thresholder.smoothImage(imagename);
				textArea.append(" filter succesfully applied.\n");
				textArea.append("IMAGE SUCCESFULLY PROCESSED");
			}

			demo_state++;

			if (demo_state == 11)
			{
				demonstrateEnd();
			}
		}
	}


	/***************************************************************************/


	class FindSeedAction extends AbstractAction
	// Identify a seed region. Display seed region in red on original image.
	{
		public void actionPerformed(ActionEvent event)
		{
			textArea.setText("");
			boolean[][] dummy = thresholder.findSeed(imagename);
		}
	}


	/***************************************************************************/


	class GrowSeedAction extends AbstractAction
	// Identify a seed region and grow it. Display region in red on original image.
	{
		public void actionPerformed(ActionEvent event)
		{
			textArea.setText("");
			thresholder.growRegion(imagename);
		}
	}


	/***************************************************************************/


	class DetectEdgeFragmentsAction extends AbstractAction
	// Identify edge fragments in the image.
	{
		public void actionPerformed(ActionEvent event)
		{
			textArea.setText("");
			thresholder.refreshImage(imagename);
			thresholder.detectEdgeFragments();
		}
	}


	/***************************************************************************/


	class DetectRegionEdgeFragmentsAction extends AbstractAction
	// Identify edge fragments in the image.
	{
		public void actionPerformed(ActionEvent event)
		{
			
			
			if (demo_state == 0)
			{
				textArea.setText("PROCESSING " + imagename + ":\n");
				textArea.append("- Reloading the image... ");
				thresholder.refreshImage(imagename);
				textArea.append(" image succesfully reloaded.\n");
				demonstrateEdgeDetectionBegin();
			}
			if (demo_state == 1)
			{
				textArea.append("- Growing a region... ");
				object_map = thresholder.detectRegionEdgesFirstStep(imagename);
				textArea.append(" region growing succesful.\n");
			}
			if (demo_state == 2)
			{
				textArea.append("- Applying an edge detector... ");
				edge_map = thresholder.detectRegionEdgesSecondStep(object_map);
				textArea.append(" edge succesfully detected.\n");
			}
			if (demo_state == 3)
			{
				textArea.append("- Refining the edge... ");
				thresholder.detectRegionEdgesThirdStep(edge_map);
				textArea.append(" edge succesfully refined.\n");
				textArea.append("IMAGE SUCCESFULLY PROCESSED");
			}

			demo_state++;

			if (demo_state == 4)
			{
				demonstrateEnd();
			}
		}
	}
}