/*
Cory McKay

NoiseRemover Class

- Objects of this class are created to remove noise from images.
- NOTE: It is assumed that the image to be manipulated is a binary image. If it isn't, the constructor
  will convert it to one using the mid brightness level as the threshold. This should not have to be
  done.
- A new NoiseRemover object should be created for each new image to be manipulated.
- The getSmoothedImage method can be called to apply a mean filter to the image. This helps to smooth
  out the image and remove noise.
- The getColourOfBackground method can be called to get the colour of the background of the image (black
  or white).
- The getNumberWhiteExpansions method can be called to get an estimate of the number of times that white
  objects in the object should be grown and then shrunk to eliminate black noise without distorting
  the image too much.
- The getNumberBlackExpansions method can be called to get an estimate of the number of times that white
  objects in the object should be grown and then shrunk to eliminate black noise without distorting
  the image too much.
- The whiteExpand method can be called to turn all black pixels with a white neighbour to white.
- The blackExpand method can be called to turn all white pixels with a black beighbour to black.
- The getWhiteExpandedShrunkImage method can be called to expand and then shrink white objects in the image
  a specified number of times. This helps to remove holes in white objects and it removes black noise.
- The getBlackExpandedShrunkImage method can be called to expand and then shrink black objects in the image
  a specified number of times. This helps to remove white noise and it removes holes in black objects.
*/

import java.awt.image.*;
import java.awt.*;

class  NoiseRemover
{
	// Fields
	private BufferedImage image; // the image that is being manipulated
	private int width;           // the width of image
	private int height;          // the height of image
	private char[][] pixels;     // an array that is used to store the alterations being made to
	                             //    image before they are applied. 'b' represents black and
	                             //    'w' represents white.


	/***************************************************************************/


	public NoiseRemover (BufferedImage base_image)
	// Constructor
	// Takes in the image that will be manipulated, calculates its width and height
	// and sets up the pixels array to mirror the contents of the image. Everything
	// that is on the dark half of the intensity spectrum is considered black and
	// everything that is on the light half is considered white for the purposes of
	// the pixels array.
	{
		image = base_image;
		width = image.getWidth();
		height = image.getHeight();
		pixels = new char[width][height];


		// Set up the pixels array
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (findPixelBrightness(image.getRGB(x,y)) < 128)
					pixels[x][y] = 'b';
				else
					pixels[x][y] = 'w';
			}
		}
	}


	/***************************************************************************/


	public BufferedImage getSmoothedImage()
	// Apply a mean filter to the image and return the altered image.
	// A 3x3 equally weighted window is used.
	// This helps to smooth out the image and remove noise.
	{
		char[][] smoothed_pixels = new char[width][height]; // temporarily changes modifications
		int total; // the total number of white pixels in the 8-neighbourhood of a pixel
		           // (including the pixel)
		int number_pixels_considered; // the number of pixels considered when averaging a pixel
		                              // and its neighbourhood (needed to make border pixels work)
		double average; // the average number of white pixels per pixel in a pixel's 8-neighbourhood
		                //  (including the pixel)


		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				total = 0;
				number_pixels_considered = 0;

				// Count the number of white pixels in the 8-neighbourhood of a pixel, being careful
				// about the borders of the image
				if ( x != 0 && y != 0 )
				{
					number_pixels_considered++;
					if (pixels[x-1][y-1] == 'w')
						total++;
				}
				if ( x != 0 )
				{
					number_pixels_considered++;
					if (pixels[x-1][y] == 'w')
						total++;
				}
				if ( x != 0 && y != height - 1 )
				{
					number_pixels_considered++;
					if (pixels[x-1][y+1] == 'w')
						total++;
				}
				if ( y != 0 )
				{
					number_pixels_considered++;
					if (pixels[x][y-1] == 'w')
						total++;
				}
				number_pixels_considered++;
				if (pixels[x][y] == 'w')
					total++;
				if ( y != height - 1 )
				{
					number_pixels_considered++;
					if (pixels[x][y+1] == 'w')
						total++;
				}
				if ( x != width -1 && y != 0 )
				{
					number_pixels_considered++;
					if (pixels[x+1][y-1] == 'w')
						total++;
				}
				if ( x != width - 1 )
				{
					number_pixels_considered++;
					if (pixels[x+1][y] == 'w')
						total++;
				}
				if ( x != width - 1 && y != height - 1 )
				{
					number_pixels_considered++;
					if (pixels[x+1][y+1] == 'w')
						total++;
				}


				// Find out if there are more white pixels or black pixels
				average = (double) total / (double) number_pixels_considered;


				// Change the smoothed_pixels array to reflect the averaged out pixel values
				if ( average > 0.5 )
					smoothed_pixels[x][y] = 'w';
				else
					smoothed_pixels[x][y] = 'b';
			}
		}

		// Return the smoothed image
		pixels = smoothed_pixels;
		applyChangesToImage();
		return image;
	}


	/***************************************************************************/


	public char getColourOfBackground()
	// Return the colour of the background of the image. This is determined by
	// counting the total number of white pixels and the total number of black pixels.
	// It is assumed that the colour with the larger number of pixels is the background
	// colour. Black is returned as 'b' and white is returned as 'w'.
	{
		int white = 0; // the number of white pixels
		int black = 0; // the number of black pixels


		// Count the total number of white and black pixels
		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				if (pixels[x][y] == 'b')
					black++;
				else if (pixels[x][y] == 'w')
					white++;
			}
		}


		// Return the colour of the background
		if (black > white)
			return 'b';
		else
			return 'w';
	}


	/***************************************************************************/


	public int getNumberWhiteExpansions()
	// Return an estimate of the number of times that white objects in the object
	// should be grown and then shrunk to eliminate black noise without distorting
	// the image too much.
	{
		ComponentLabeler counter = new ComponentLabeler(pixels, width, height);
		return counter.getEstimatedWhiteGrowths();
	}


	/***************************************************************************/


	public int getNumberBlackExpansions()
	// Return an estimate of the number of times that black objects in the object
	// should be grown and then shrunk to eliminate white noise without distorting
	// the image too much.
	{
		ComponentLabeler counter = new ComponentLabeler(pixels, width, height);
		return counter.getEstimatedBlackGrowths();
	}


	/***************************************************************************/


	public BufferedImage whiteExpand()
	// Turn all black pixels with a white neighbour to white and return the image
	{
		pixels = expand();
		applyChangesToImage();
		return image;
	}


	/***************************************************************************/


	public BufferedImage blackExpand()
	// Turn all white pixels with a black neighbour to black and return the image
	{
		pixels = shrink();
		applyChangesToImage();
		return image;
	}


	/***************************************************************************/


	public BufferedImage getWhiteExpandedShrunkImage(int times_to_expand)
	// Expand and then shrink white objects in the image times_to_expand times and
	// return the modified image.
	// This helps to remove holes in white objects and it removes black noise.
	{
		// Expand and shrink the image, recording the changes in the pixels array
		for (int i = 0; i < times_to_expand; i++)
			pixels = expand();
		for (int i = 0; i < times_to_expand; i++)
			pixels = shrink();


		// Make image reflect the expansion and shrinking and return the image
		applyChangesToImage();
		return image;
	}


	/***************************************************************************/


	public BufferedImage getBlackExpandedShrunkImage(int times_to_shrink)
	// Expand and then shrink black objects in the image times_to_shrink times and
	// return the modified image.
	// This helps to remove white noise and it removes holes in black objects.
	{
		// Shrink and exapand the image, recording the changes in the pixels aray
		for (int i = 0; i < times_to_shrink; i++)
			pixels = shrink();
		for (int i = 0; i < times_to_shrink; i++)
			pixels = expand();


		// Make image reflect the expansion and shrinking and return the image
		applyChangesToImage();
		return image;
	}


	/***************************************************************************/


	private char[][] expand()
	// Change each pixel from black to white if any 8-neighbours of the pixel are white
	{
		boolean found_one; // is set to true if a black pixel has a white neighbour
		char[][] expanded_pixels = new char[width][height]; // reflects the image after expansion

		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				found_one = false;


				// Check to see if neighbours of black points are white, being careful not to
				// exceed array boundaries for pixels on the edge of the image
				if (pixels[x][y] == 'b')
				{
					if ( x != 0 && y != 0 )
					{
						if (pixels[x-1][y-1] == 'w')
							found_one = true;
					}
					if ( x != 0 )
					{
						if (pixels[x-1][y] == 'w')
							found_one = true;
					}
					if ( x != 0 && y != height - 1 )
					{
						if (pixels[x-1][y+1] == 'w')
							found_one = true;
					}
					if ( y != 0 )
					{
						if (pixels[x][y-1] == 'w')
							found_one = true;
					}
					if ( y != height - 1 )
					{
						if (pixels[x][y+1] == 'w')
							found_one = true;
					}
					if ( x != width -1 && y != 0 )
					{
						if (pixels[x+1][y-1] == 'w')
							found_one = true;
					}
					if ( x != width - 1 )
					{
						if (pixels[x+1][y] == 'w')
							found_one = true;
					}
					if ( x != width - 1 && y != height - 1 )
					{
						if (pixels[x+1][y+1] == 'w')
							found_one = true;
					}
				}


				// Mark the pixel value in expanded_pixels
				if (found_one == true)
					expanded_pixels[x][y] = 'w';
				else
					expanded_pixels[x][y] = pixels[x][y];
			}
		}


		return expanded_pixels;
	}


	/***************************************************************************/


	private char[][] shrink()
	// Change each pixel from white to black if any 8-neighbours of the pixel are black
	{
		boolean found_one; // is set to true if a white pixel has a black neighbour
		char[][] shrunken_pixels = new char[width][height]; // reflects the image after expansion


		for (int y = 0; y < height; y++)
		{
			for (int x = 0; x < width; x++)
			{
				found_one = false;

				// Check to see if neighbours of white points are black, being careful not to
				// exceed array boundaries for pixels on the edge of the image
				if (pixels[x][y] == 'w')
				{
					if ( x != 0 && y != 0 )
					{
						if (pixels[x-1][y-1] == 'b')
							found_one = true;
					}
					if ( x != 0 )
					{
						if (pixels[x-1][y] == 'b')
							found_one = true;
					}
					if ( x != 0 && y != height - 1 )
					{
						if (pixels[x-1][y+1] == 'b')
							found_one = true;
					}
					if ( y != 0 )
					{
						if (pixels[x][y-1] == 'b')
							found_one = true;
					}
					if ( y != height - 1 )
					{
						if (pixels[x][y+1] == 'b')
							found_one = true;
					}
					if ( x != width -1 && y != 0 )
					{
						if (pixels[x+1][y-1] == 'b')
							found_one = true;
					}
					if ( x != width - 1 )
					{
						if (pixels[x+1][y] == 'b')
							found_one = true;
					}
					if ( x != width - 1 && y != height - 1 )
					{
						if (pixels[x+1][y+1] == 'b')
							found_one = true;
					}
				}

				// Mark the pixel value in expanded_pixels
				if (found_one == true)
					shrunken_pixels[x][y] = 'b';
				else
					shrunken_pixels[x][y] = pixels[x][y];
			}
		}


		return shrunken_pixels;
	}


	/***************************************************************************/


	private void applyChangesToImage()
	// Change image to reflect the contents of the pixels array
	{
		// Prepare to write to image
		WritableRaster raster = image.getRaster();
		ColorModel model = image.getColorModel();
		Color white = Color.white;
		Color black = Color.black;
		int argbWhite = white.getRGB();
		int argbBlack = black.getRGB();
		Object colorDataWhite = model.getDataElements(argbWhite, null);
		Object colorDataBlack = model.getDataElements(argbBlack, null);


		// Convert image to reflect the contenst of pixels
		for (int y = 0; y < height; y++)
		{
			for (int x= 0; x < width; x++)
			{
				if (pixels[x][y] == 'w')
					raster.setDataElements(x, y, colorDataWhite);
				else
					raster.setDataElements(x, y, colorDataBlack);
			}
		}
	}


	/***************************************************************************/


	private int findPixelBrightness(int pixel_data)
	// Returns the brightness of the pixel given in pixel_data
	{
		int red   = (0x00ff0000 & pixel_data) >> 16;
		int green = (0x0000ff00 & pixel_data) >> 8;
		int blue  = (0x000000ff & pixel_data);
		/* return (int)(.33*red + .56*green + .11*blue); THIS IS HUMAN BALANCED */
		return (int)(.333*red + .333*green + .333*blue);
	}
}