/*
Cory McKay

ThresholdCalculator Class

- Objects of this class are used to automatically determine the appropriate
  threshold to bring out an object in an image with the given histogram.
- The getThreshold method can be called to get the calculated
  appropriate threshold for the specified histogram.
- The findAllPeaks method can be called to get an array containing the indices
  of all significant local maxima of the specified histogram.
*/

class ThresholdCalculator 
{
	// Fields
	private int hist[]; // the brightness level histogram for the image. The pixels are
	                    // given a brightness rating from 0 (dark) to 255 (light).
	                    // The historgram records the frequency of each of these brightness
	                    // levels in the image that is being processed.

	
	/***************************************************************************/


	public ThresholdCalculator(int histogram[])
	// Constructor
	// histogram is a brightness histogram that will be used to perform calculations
	{
		hist = histogram;
	}


	/***************************************************************************/


	public int getThreshold()
	// Returns the value of the threshold that is calculated to be the best for bringing
	// out an object. This is done by calculating peakiness, as defined by the mode method
	// for finding an appropriate threshold.
	// The following two conditions are also imposed in order to improve results:
	//    - to be considered, any peaks must contain at least 1/1000 of the total number of
	//      pixels in the image
	//    - valleys must be 1/3 or less of the highest level in the histogram
	// If no acceptable threshold is found because of the imposition of these two conditions,
	// then they are not applied, and the simple mode method is used to find the threshold.
	{
		int[] peaks = findAllPeaks(); // indices of all significant peaks in the image
		int valley_indice; // the indice of a valley between two peaks
		int valley_value; // the value on the histogram of a valley
		int numerator; // the numerator used in the peakiness calculation
		double peakiness; // the peakiness, as defined by the mode method
		PeakinessList list = new PeakinessList(); // a list of valley indices and
		                                          // related peakinesses
		int max_hist = 0; // the most common brightness range in the histogram
		int number_pixels = 0; // the total number of pixels in the image


		// Calculate max_hist
		for (int i = 0; i < 256; i++)
		{
			if (hist[i] > max_hist)
				max_hist = hist[i];
		}


		// Calculate number_pixels
		for (int i = 0; i < 256; i++)
		{
			number_pixels = number_pixels + hist[i];
		}


		// Add valley indices and peakiness values for all local maxima pairs to list.
		// Peakiness is calculated using the formla:
		//		peakiness = min(left_peak_value, right_peak_value) / valley_value
		// The following two conditions are also imposed in order to improve results:
		//    - to be considered any peaks must contain at least 1/1000 of the total number of
		//      pixels in the image
		//    - valleys must be 1/3 or less of the highest level in the histogram
		for (int i = 0; i < peaks.length; i++)
		{
			for (int j = i+1; j < peaks.length; j++)
			{
				if ( ( hist[ peaks[i] ] > number_pixels / 1000 ) && ( hist[ peaks[j] ] > number_pixels / 1000 ) )
				{
					valley_indice = findValley( peaks[i], peaks[j] );
					valley_value = hist[ valley_indice ];
					if ( valley_value < max_hist / 3)
					{
						if (valley_value == 0) // avoid division by 1
							valley_value = 1;
						numerator = Math.min( hist[ peaks[j] ], hist[ peaks[i] ] );
						peakiness = (double) numerator / (double) valley_value;
						list.insert_new_node(valley_indice, peakiness);
					}
				}
			}
		}


		// If no candidate threshold were found by the previous method, redo list using
		// the basic mode method
		if (list.get_number_of_nodes() == 0)
		{
			list = new PeakinessList();
			for (int i = 0; i < peaks.length; i++)
			{
				for (int j = i+1; j < peaks.length; j++)
				{
					valley_indice = findValley( peaks[i], peaks[j] );
					valley_value = hist[ valley_indice ];
					if (valley_value == 0) // avoid division by 1
						valley_value = 1;
					numerator = Math.min( hist[ peaks[j] ], hist[ peaks[i] ] );
					peakiness = (double) numerator / (double) valley_value;
					list.insert_new_node(valley_indice, peakiness);
				}
			}
		}


		// Return the index of the valley that corresponds to the highest peakiness value
		PeakinessObject[] results = list.get_peakiness_objects();
		PeakinessObject best = results[0];
		for (int i = 1; i < list.get_number_of_nodes(); i++)
		{
			if (results[i].peakiness > best.peakiness)
				best = results[i];
		}
		return best.valley_indice;
	}


	/***************************************************************************/


	public int[] findAllPeaks()
	// Returns an array of integers that contains the histogram indices of all significant
	// local maxima on the brightness histogram.
	// To be considered a significant local maxima, the indices must have lower
	// values immediately to their left an right (except for on the extreme right
	// and extreme left), and there must not be any other higher local maxima within
	// less than 7 units to the right or left of the indice.
	{
		PeakList peaks = new PeakList(); // a PeakList that will hold all peaks found (even
		                                 // if there is another local maxima within less than
		                                 //7 units
		int[] results; // the indices of all the peaks placed in the peaks list
		int size; // the number of peaks in the results array
		int[] filtered_results; // the peaks in the results array, minus those peaks that
		                        // are within less than 7 units of a higher peak
		int size_after_filtering; // the number of peaks in the filtered_results array
		boolean too_close; // true if two peaks are within less than 7 units, false otherwise
		int new_peak; // the indice of an individual peak

		
		// Put all local maxima in the peaks list
		new_peak = findPeak(Integer.MAX_VALUE);
		while (new_peak != -1)
		{
			peaks.insert_new_peak(new_peak);
			new_peak = findPeak( hist[new_peak] );
		}


		// Transfer the contents of the peaks list into the results array and set the values
		// of size and size_after_filtering
		size = peaks.get_number_of_peaks();
		size_after_filtering = size;
		results = peaks.get_peak_indices();

		
		// Discount all peaks that are less than 7 units away from a higher peak already 
		// encountered by setting their values in the results array to -1
		for (int i = 0; i < size; i++)
		{
			too_close = false;
			for (int j = 0; j < i; j++)
			{
				if ( (Math.abs(results[i] - results[j]) < 7)  && (results[j] != -1) )
					too_close = true;
			}	
			if (too_close == true)
			{
				results[i] = -1;
				size_after_filtering--;
			}
		}


		// Fill the filtered_results array by placing all non -1 values of the results
		// array in it
		filtered_results = new int[size_after_filtering];
		int j = 0;
		for (int i =0; i < size; i++)
		{
			if (results[i] != -1)
			{
				filtered_results[j] = results [i];
				j++;
			}
		}


		/* FOR TESTING PURPOSES
		// Print out the size and the contents of the filtered_results array
		System.out.println("-> " + size_after_filtering);
		for (int i = 0; i<size_after_filtering; i++)
		{
				System.out.print(filtered_results[i] + " ");
		}
		System.out.println();
		*/

		// Return the indices of the significant peaks
		return filtered_results;
	}


	/***************************************************************************/


	private int findPeak(int vertical_limit)
	// Returns the index of the value in the histogram with the highest value that is 
	// below vertical_limit and is a local maximum (the values to the left and right
	// are lower than its value). Returns -1 if no such peak can be found.
	{
		int highest_so_far; // the highest peak found by this method so far
		int peak_location; // the indice of the highest peak found so far by this method
		

		// Count the leftmost value of the histogram if the point to the right is
		// below it and it is below vertical_limit
		if ( (hist[0] < vertical_limit) && (hist[1] < hist[0]) )
		{
			highest_so_far = hist[0];
			peak_location = 0;
		}
		else
		{
			highest_so_far = 0;
			peak_location = -1;
		}


		// Find the highest local maximum below vertical_limit
		for (int i = 1; i < 256; i++)
		{
			// Special case of the rightmost value on the histogram
			if ( (i == 255) && (hist[i] > highest_so_far) && (hist[i] < vertical_limit) )
				peak_location = 255;
			else if ( (hist[i] > highest_so_far) && (hist[i] < vertical_limit) &&
				 (hist[i-1] < hist[i]) && (hist[i+1] < hist[i]) )
			{
				highest_so_far = hist[i];
				peak_location = i;
			}
		}
		
		
		// Return the indice of the highest acceptable peak
		return peak_location;
	}


	/***************************************************************************/


	private int findValley(int first_limit, int second_limit)
	// Returns the lowest point in the histogram between indices first_limit and
	// second_limit
	{
		// Reverse first_limit and second_limit if first_limit is to the right
		// of second_limit
		if (first_limit > second_limit)
		{
			int temp = first_limit;
			first_limit = second_limit;
			second_limit = temp;
		}
		

		// Find the lowest point within the given limits
		int lowest_so_far = hist[first_limit]; // the lowest point found so far
		int valley_location = first_limit; // the indice of the lowest point found
		                                   // so far
		for (int i = first_limit; i <= second_limit; i++)
		{
			if (hist[i] < lowest_so_far)
			{
				lowest_so_far = hist[i];
				valley_location = i;
			}
		}
		
		
		// Return the indice of the lowest point between the given limits
		return valley_location;
	}
}